@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2011	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -112,6 +112,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 	POOL_STATUS status;
 	char *string;
 	int lock_kind;
+	int specific_error = 0;
 
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
@@ -206,6 +207,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			{
 				pool_query_context_destroy(query_context);
 				pool_set_skip_reading_from_backends();
+				pool_memory = old_context;
 				return POOL_CONTINUE;
 			}
 		}
@@ -215,6 +217,18 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		 */
 		pool_start_query(query_context, contents, len, node);
 
+		/*
+		 * Check if multi statement query
+		 */
+		if (parse_tree_list && list_length(parse_tree_list) > 1)
+		{
+			query_context->is_multi_statement = true;
+		}
+		else
+		{
+			query_context->is_multi_statement = false;
+		}
+
 		if (PARALLEL_MODE)
 		{
 			bool parallel = true;
@@ -224,6 +238,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			if (parallel)
 			{
 				pool_query_context_destroy(query_context);
+				pool_memory = old_context;
 				return status;
 			}
 		}
@@ -277,6 +292,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 
 			pool_query_context_destroy(query_context);
 			pool_set_skip_reading_from_backends();
+			pool_memory = old_context;
 			return POOL_CONTINUE;
 		}
 
@@ -414,14 +430,13 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 				rewrite_query = rewrite_timestamp(backend, query_context->parse_tree, false, msg);
 
 				/*
-				 * If the query is BEGIN READ WRITE in master/slave mode,
-				 * we send BEGIN instead of it to slaves/standbys.
+				 * If the query is BEGIN READ WRITE or
+				 * BEGIN ... SERIALIZABLE in master/slave mode,
+				 * we send BEGIN to slaves/standbys instead.
 				 * original_query which is BEGIN READ WRITE is sent to primary.
 				 * rewritten_query which is BEGIN is sent to standbys.
 				 */
-				if (is_start_transaction_query(query_context->parse_tree) &&
-					is_read_write((TransactionStmt *)query_context->parse_tree) &&
-					MASTER_SLAVE)
+				if (pool_need_to_treat_as_if_default_transaction(query_context))
 				{
 					rewrite_query = pstrdup("BEGIN");
 				}
@@ -452,15 +467,39 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 				pool_query_context_destroy(query_context);
 				return POOL_END;
 			}
+
+			/* Check specific errors */
+			specific_error = check_errors(backend, MASTER_NODE_ID);
+			if (specific_error)
+			{
+				/* log error message */
+				generate_error_message("SimpleQuery: ", specific_error, contents);
+			}
 		}
 
-		/*
-		 * Send the query to other than master node.
-		 */
-		if (pool_send_and_wait(query_context, -1, MASTER_NODE_ID) != POOL_CONTINUE)
+		if (specific_error)
 		{
-			pool_query_context_destroy(query_context);
-			return POOL_END;
+			char msg[1024] = POOL_ERROR_QUERY; /* large enough */
+			int len = strlen(msg);
+
+			memset(msg + len, 0, sizeof(int));
+
+			/* send query to other nodes */
+			query_context->rewritten_query = msg;
+			query_context->rewritten_length = len;
+			if (pool_send_and_wait(query_context, -1, MASTER_NODE_ID) != POOL_CONTINUE)
+				return POOL_END;
+		}
+		else
+		{
+			/*
+			 * Send the query to other than master node.
+			 */
+			if (pool_send_and_wait(query_context, -1, MASTER_NODE_ID) != POOL_CONTINUE)
+			{
+				pool_query_context_destroy(query_context);
+				return POOL_END;
+			}
 		}
 
 		/* Send "COMMIT" or "ROLLBACK" to only master node if query is "COMMIT" or "ROLLBACK" */
@@ -706,8 +745,6 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			return POOL_END;
 		}
 
-		session_context->uncompleted_message = msg;
-
 		/*
 		 * Decide where to send query
 		 */
@@ -820,6 +857,12 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		}
 	}
 
+	/*
+	 * need to set uncompleted_message after calling ReadyForQuery(),
+	 * because the function destroys query context of uncompleted_message.
+	 */
+	session_context->uncompleted_message = msg;
+
 	/*
 	 * Cannot call free_parser() here. Since "string" might be allocated in parser context.
 	 * free_parser();
@@ -1328,7 +1371,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 
 			TSTATE(backend, i) = kind;
 
-			pool_debug("ReadyForQuery: transaction state:%c", state);
+			pool_debug("ReadyForQuery: transaction state:%c", kind);
 
 			/*
 			 * The transaction state to be returned to frontend is
@@ -1381,7 +1424,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 				 * SESSION CHARACTERISTICS AS TRANSACTION ISOLATION LEVEL
 				 * SERIALIZABLE, remember it.
 				 */
-				else if (is_set_transaction_serializable(node, query))
+				else if (is_set_transaction_serializable(node))
 				{
 					pool_set_transaction_isolation(POOL_SERIALIZABLE);
 				}
@@ -1411,10 +1454,10 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 		pool_unset_query_in_progress();
 	}
 
-	if (!pool_is_doing_extended_query_message())
+	if (session_context->query_context && !pool_is_doing_extended_query_message())
 	{
 		if (!(node && IsA(node, PrepareStmt)))
-			pool_query_context_destroy(pool_get_session_context()->query_context);
+			pool_query_context_destroy(session_context->query_context);
 	}
 
 	sp = MASTER_CONNECTION(backend)->sp;
@@ -1779,6 +1822,20 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 			p1 = p2;
 			len1 = strlen(p2) + 1;
 		}
+		/*
+		 * If PREPARE or extended query protocol commands caused error,
+		 * remove the temporary saved message.
+		 */
+		else
+		{
+			if (session_context->uncompleted_message)
+			{
+				pool_add_sent_message(session_context->uncompleted_message);
+				pool_remove_sent_message(session_context->uncompleted_message->kind,
+										 session_context->uncompleted_message->name);
+				session_context->uncompleted_message = NULL;
+			}
+		}
 
 		pool_write(frontend, "C", 1);
 		sendlen = htonl(len1+4);
@@ -1801,7 +1858,6 @@ POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 						   POOL_CONNECTION_POOL *backend)
 {
 	POOL_STATUS ret;
-	POOL_SESSION_CONTEXT *session_context;
 
 	ret = SimpleForwardToFrontend('E', frontend, backend);
 	if (ret != POOL_CONTINUE)
@@ -1866,21 +1922,6 @@ POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 	}
 #endif
 
-	/* An error occurred with PREPARE or DEALLOCATE command.
-	 * Free pending portal object.
-	 */
-	session_context = pool_get_session_context();
-	if (session_context)
-	{
-		if (session_context->uncompleted_message)
-		{
-			pool_add_sent_message(session_context->uncompleted_message);
-			pool_remove_sent_message(session_context->uncompleted_message->kind,
-									 session_context->uncompleted_message->name);
-			session_context->uncompleted_message = NULL;
-		}
-	}
-
 	return POOL_CONTINUE;
 }
 
@@ -2027,7 +2068,15 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 		return POOL_END;
 
 	if (fkind != 'S' && pool_is_ignore_till_sync())
+	{
+		/*
+		 * Flag setting for calling ProcessBackendResponse()
+		 * in pool_process_query().
+		 */
+		if (!pool_is_query_in_progress())
+			pool_set_query_in_progress();
 		return POOL_CONTINUE;
+	}
 
 	pool_unset_doing_extended_query_message();
 
@@ -2037,6 +2086,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 		char *query;
 		Node *node;
 		List *parse_tree_list;
+		POOL_MEMORY_POOL *old_context = pool_memory;
 
 		case 'X':	/* Terminate */
 			return POOL_END;
@@ -2102,6 +2152,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 				return POOL_END;
 			}
 			query = "INSERT INTO foo VALUES(1)";
+			pool_memory = query_context->memory_context;
 			parse_tree_list = raw_parser(query);
 			node = (Node *) lfirst(list_head(parse_tree_list));
 			pool_start_query(query_context, query, strlen(query) + 1, node);
@@ -2112,6 +2163,8 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 				status = FunctionCall3(frontend, backend, len, contents);
 			else
 				status = FunctionCall(frontend, backend);
+
+			pool_memory = old_context;
 			break;
 
 		case 'c':	/* CopyDone */
@@ -2152,6 +2205,8 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 
 	if (pool_is_ignore_till_sync())
 	{
+		if (pool_is_query_in_progress())
+			pool_unset_query_in_progress();
 		return POOL_CONTINUE;
 	}
 
@@ -2753,7 +2808,7 @@ static int check_errors(POOL_CONNECTION_POOL *backend, int backend_id)
 	 * M:S2:ERROR:  could not serialize access due to concurrent update
 	 * S:S2:UPDATE t1 SET i = i + 1; <-- success in UPDATE and data becomes inconsistent!
 	 */
-	if (detect_serialization_error(CONNECTION(backend, backend_id), MAJOR(backend)) == SPECIFIED_ERROR)
+	if (detect_serialization_error(CONNECTION(backend, backend_id), MAJOR(backend), true) == SPECIFIED_ERROR)
 		return 2;
 
 	/*
