@@ -6,7 +6,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2009	PgPool Global Development Group
+ * Copyright (c) 2003-2011	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -33,6 +33,7 @@
 #include "parser/parsenodes.h"
 #include "pool_rewrite_query.h"
 #include "pool_session_context.h"
+#include "pool_process_reporting.h"
 
 #define SPECIFIED_ERROR 1
 #define POOL_DUMMY_WRITE_QUERY "DELETE FROM foo WHERE col = 'pgpool: unable to parse the query'"
@@ -95,9 +96,6 @@ extern POOL_STATUS CloseComplete(POOL_CONNECTION *frontend,
 extern POOL_STATUS CommandComplete(POOL_CONNECTION *frontend,
 								   POOL_CONNECTION_POOL *backend);
 
-extern POOL_STATUS ParameterDescription(POOL_CONNECTION *frontend,
-										POOL_CONNECTION_POOL *backend);
-
 extern POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 								  POOL_CONNECTION_POOL *backend);
 
@@ -153,10 +151,10 @@ extern int RowDescription(POOL_CONNECTION *frontend,
 
 extern POOL_STATUS wait_for_query_response(POOL_CONNECTION *frontend, POOL_CONNECTION *backend, int protoVersion);
 extern int is_select_query(Node *node, char *sql);
-extern int is_sequence_query(Node *node);
-extern int is_start_transaction_query(Node *node);
-extern int is_commit_query(Node *node);
-extern int is_strict_query(Node *node); /* returns non 0 if this is strict query */
+extern bool is_commit_query(Node *node);
+extern bool is_rollback_query(Node *node);
+extern bool is_commit_or_rollback_query(Node *node);
+extern bool is_strict_query(Node *node); /* returns non 0 if this is strict query */
 extern int need_insert_lock(POOL_CONNECTION_POOL *backend, char *query, Node *node);
 extern POOL_STATUS insert_lock(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend, char *query, InsertStmt *node, int lock_kind);
 extern char *parse_copy_data(char *buf, int len, char delimiter, int col_id);
@@ -165,19 +163,13 @@ extern void query_ps_status(char *query, POOL_CONNECTION_POOL *backend);		/* sho
 extern POOL_STATUS start_internal_transaction(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend, Node *node);
 extern POOL_STATUS end_internal_transaction(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 extern int detect_deadlock_error(POOL_CONNECTION *master, int major);
-extern int detect_serialization_error(POOL_CONNECTION *master, int major);
+extern int detect_serialization_error(POOL_CONNECTION *master, int major, bool unread);
 extern int detect_active_sql_transaction_error(POOL_CONNECTION *backend, int major);
 extern int detect_query_cancel_error(POOL_CONNECTION *backend, int major);
 extern bool is_partition_table(POOL_CONNECTION_POOL *backend, Node *node);
 extern POOL_STATUS pool_discard_packet(POOL_CONNECTION_POOL *cp);
 extern void query_cache_register(char kind, POOL_CONNECTION *frontend, char *database, char *data, int data_len);
-
 extern int is_drop_database(Node *node);		/* returns non 0 if this is a DROP DATABASE command */
-extern void config_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
-extern void pools_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
-extern void processes_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
-extern void nodes_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
-extern void version_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 
 extern POOL_STATUS send_simplequery_message(POOL_CONNECTION *backend, int len, char *string, int major);
 extern POOL_STATUS send_extended_protocol_message(POOL_CONNECTION_POOL *backend,
